import asyncio
import json
import typing as t
from urllib.parse import urljoin
import signal
from xxlimited import Str

import aiohttp

from . import exceptions


class Response(t.NamedTuple):
    content: bytes
    status: int


def _exit_handler(
    sig_num: 'signal.Signals',
    flag: t.List[bool],
    consumer: 'KafkaBridgeConsumer'
) -> None:
    if flag:
        raise SystemExit(128 + sig_num)
    else:
        loop = asyncio.get_running_loop()
        loop.run_until_complete(consumer.delete_consumer_instance())
        flag.append(True)


def _graceful_exit(consumer: 'KafkaBridgeConsumer') -> None:
    flag: t.List[bool] = []
    loop = asyncio.get_running_loop()
    for sig_num in (signal.SIGINT, signal.SIGTERM):
        loop.add_signal_handler(sig_num, _exit_handler, sig_num, consumer, flag)


class KafkaBridgeConsumer:
    """
    Docs: https://strimzi.io/docs/bridge/latest
    Example: https://strimzi.io/blog/2019/11/05/exposing-http-bridge/
    """
    TOPICS_PATH = 'topics'
    CONSUMER_PATH = 'consumers/{group_id}'
    CONSUMER_INSTANCE_PATH = 'consumers/{group_id}/instances/{name}'
    SUBSCRIPTION_PATH = 'consumers/{group_id}/instances/{name}/subscription'
    RECORDS_PATH = 'consumers/{group_id}/instances/{name}/records'
    COMMIT_PATH = 'consumers/{group_id}/instances/{name}/offsets'

    def __init__(
        self,
        *topics: str,
        bootstrap_server: str,
        group_id: str,
        auto_offset_reset: str,
        # auto_offset_reset: t.Literal['earliest', 'latest'],
        enable_auto_commit: bool,
        consumer_name: str,
        sleep_interval_seconds: int = 2,
        client_timeout_seconds: int = 15,
        headers: t.Dict[str, t.Any] = None,
        proxy: str = 'strimzi',
        # proxy: t.Literal['strimzi', 'confluent'] = 'strimzi',
        content_type: str = 'application/vnd.kafka.v2+json'
    ) -> None:
        self._content_type = content_type
        self._group_id = group_id
        self._consumer_name = consumer_name
        self._topics = topics
        self._offsets: t.Dict[t.Tuple[str, str], t.Dict[str, t.Any]] = {}
        self._proxy = proxy
        if proxy == 'strimzi':
            self._config = {
                'auto.offset.reset': auto_offset_reset,
                'enable.auto.commit': enable_auto_commit,
                'format': 'binary',
                'name': consumer_name,
            }
        elif proxy == 'confluent':
            self._config = {
                'auto.offset.reset': auto_offset_reset,
                'auto.commit.enable': 'true' if enable_auto_commit else 'false',
                'format': 'binary',
                'name': consumer_name,
            }

        # Delay between fetching of records if
        # the previous fetch return zero records
        self._sleep_interval_seconds = sleep_interval_seconds
        self._bootstrap_server = bootstrap_server
        self._client_timeout_seconds = client_timeout_seconds
        self._headers = headers or {}

    @property
    def is_strimzi_proxy(self) -> bool:
        return self._proxy == 'strimzi'

    async def _request(
        self,
        method: str,
        path: str,
        *,
        params: t.Dict[str, t.Any] = None,
        data: t.Dict[str, t.Any] = None,
        headers: t.Dict[str, t.Any] = None,
    ) -> Response:
        params = params or {}
        data = data or {}
        _headers = headers or {}
        _headers.update(self._headers)
        _headers['Content-Type'] = self._content_type
        url = urljoin(self._bootstrap_server, path)

        async with aiohttp.ClientSession(
            timeout=aiohttp.ClientTimeout(self._client_timeout_seconds),
            headers=_headers,
        ) as session:
            try:
                resp = await session.request(
                    method=method,
                    url=url,
                    params=params,
                    json=data,
                )
            except aiohttp.ClientError as e:
                raise exceptions.KafkaBridgeClientError(e)
            except asyncio.TimeoutError as e:
                raise exceptions.KafkaBridgeTimeoutError(e)
            else:
                content = await resp.read()
                return Response(content=content, status=resp.status)

    async def __aenter__(self) -> None:
        try:
            # if process killed by OOM we cant delete consumer
            # so we need try delete consumer before creating new consumer
            await self.delete_consumer_instance()
        except exceptions.ConsumerNotFound:
            pass

        await self._create_consumer_instance()
        try:
            await self._subscribe()
        except exceptions.KafkaBridgeError:
            await self.delete_consumer_instance()
            raise

        _graceful_exit(self)

        return None

    async def __aexit__(self, *args: t.Any, **kwargs: t.Any) -> None:
        await self.delete_consumer_instance()

    async def _create_consumer_instance(self) -> None:
        response = await self._request(
            'POST',
            self.CONSUMER_PATH.format(group_id=self._group_id),
            data=self._config,
        )
        if response.status != 200:
            raise exceptions.KafkaBridgeError(
                f'status: {response.status}, text: {response.content!r}',
            )

    async def delete_consumer_instance(self) -> None:
        if not self._consumer_name:
            return None

        response = await self._request(
            'DELETE',
            self.CONSUMER_INSTANCE_PATH.format(
                group_id=self._group_id,
                name=self._consumer_name,
            ),
        )
        if response.status != 204:
            if response.status == 404:
                raise exceptions.ConsumerNotFound
            else:
                raise exceptions.KafkaBridgeError(
                    f'status: {response.status}, text: {response.content!r}',
                )

    async def _subscribe(self) -> None:
        response = await self._request(
            'POST',
            self.SUBSCRIPTION_PATH.format(
                group_id=self._group_id,
                name=self._consumer_name,
            ),
            data={'topics': self._topics},
        )

        if response.status != 204:
            raise exceptions.KafkaBridgeError(
                f'status: {response.status}, text: {response.content!r}',
            )

    async def get_records(
        self,
        *,
        max_bytes: int,
        timeout: int = 10000
    ) -> t.AsyncGenerator[t.Dict[str, t.Any], None]:
        while True:
            response = await self._request(
                'GET',
                self.RECORDS_PATH.format(
                    group_id=self._group_id,
                    name=self._consumer_name,
                ),
                params={'timeout': timeout, 'max_bytes': max_bytes},
                headers={'Accept': 'application/vnd.kafka.binary.v2+json'},
            )
            if response.status != 200:
                raise exceptions.KafkaBridgeError(
                    f'status: {response.status}, text: {response.content!r}',
                )
            records = json.loads(response.content)

            for rec in records:
                self._offsets[(rec['topic'], rec['partition'])] = {
                    'topic': rec['topic'],
                    'partition': rec['partition'],
                    'offset': rec['offset']
                }
                yield rec

            if not records:
                await asyncio.sleep(self._sleep_interval_seconds)

    async def poll_records(
        self,
        *,
        max_bytes: int,
        timeout: int = 10000
    ) -> t.List[t.Dict[str, t.Any]]:
        response = await self._request(
            'GET',
            self.RECORDS_PATH.format(
                group_id=self._group_id,
                name=self._consumer_name,
            ),
            params={'timeout': timeout, 'max_bytes': max_bytes},
            headers={'Accept': 'application/vnd.kafka.binary.v2+json'},
        )
        if response.status != 200:
            raise exceptions.KafkaBridgeError(
                f'status: {response.status}, text: {response.content!r}',
            )
        records = json.loads(response.content)

        for rec in records:
            self._offsets[(rec['topic'], rec['partition'])] = {
                'topic': rec['topic'],
                'partition': rec['partition'],
                'offset': rec['offset']
            }

        if not records:
            await asyncio.sleep(self._sleep_interval_seconds)

        return records

    async def commit(self) -> None:
        response = await self._request(
            'POST',
            self.COMMIT_PATH.format(
                group_id=self._group_id,
                name=self._consumer_name,
            ),
            data={'offsets': list(self._offsets.values())},
        )
        success_status = 204 if self.is_strimzi_proxy else 200
        if response.status != success_status:
            raise exceptions.KafkaBridgeError(
                f'status: {response.status}, text: {response.content!r}',
            )



import typing as t
import logging
import base64

import requests as r
from mypy_extensions import TypedDict

from .exceptions import (
    KafkaBridgeTopicNotFoundError,
    ValidationError,
    KafkaBridgeResourceException,
    KafkaBridgeInternalError,
)


__all__ = ['KafkaBridgeProducer', 'Message', ]


logger = logging.getLogger(__name__)


class Message(TypedDict):
    """
    The object for representation message data.
    """
    value: t.Union[str, bytes]
    key: t.Optional[str]


class KafkaBridgeProducer:
    """
    Docs: https://strimzi.io/docs/bridge/latest

    This class is abstraction over producer from strimzi bridge.

    Example of usage:

        >>> producer = KafkaBridgeProducer('http://bridge.url' timeout=5)
        >>> producer.send(Message(key='1', value='value'))

    if you need to send binary data then u can use `binary` argument in send
    method.

    Raises:
        - KafkaBridgeTopicNotFoundError - topic not found
        - ValidationError - bad format of key or value
        - KafkaBridgeInternalError - internal error of bridge service
        - KafkaBridgeResourceException - common error
    """

    def __init__(self, url: str, timeout: float) -> None:
        self._url = url
        self._timeout = timeout

    def sends(
        self,
        topic: str,
        records: t.List[Message],
        binary: bool = False,
    ) -> bool:
        """
        This method provide aproache to send messages to kafka in
        batch (send all messages and flush) mode.
        """
        data: t.Any = None

        if binary:
            data = [
                {
                    'key': (
                        r.get('key') and base64.b64encode(str(r.get('key')).encode()).decode()
                    ),
                    # we need to conver binary data to base64 for success serialization to json
                    'value':
                        base64.b64encode(r.get('value', b'')).decode(),  # type: ignore
                }
                for r in records
            ]
            headers = {'Content-Type': 'application/vnd.kafka.binary.v2+json'}
        else:
            data = records
            headers = {'Content-Type': 'application/vnd.kafka.json.v2+json'}

        response = r.post(
            url=f'{self._url}/topics/{topic}',
            json={'records': data},
            timeout=self._timeout,
            headers=headers,
        )

        if response.status_code == 200:
            return True
        elif response.status_code == 404:
            raise KafkaBridgeTopicNotFoundError(
                str({'topic': topic, 'values': data})
            )
        elif response.status_code == 422:
            raise ValidationError(
                str({'topic': topic, 'values': data})
            )
        elif response.status_code == 500:
            raise KafkaBridgeInternalError(
                str({'topic': topic, 'values': data})
            )
        else:
            msg = (
                '[KAFKA BRIDGE] request failed: '
                'status=%(status)s data=%(data)s'
            )
            ctx = {
                'status': response.status_code,
                'data': response.text,
            }

            logger.error(msg, ctx)

            raise KafkaBridgeResourceException(
                f'Request to Kafka Bridge failed. '
                f'{str({"topic": topic})}',
            )

    def send(
        self,
        topic: str,
        record: Message,
        binary: bool = False
    ) -> bool:
        """
        This method provide aproache to send messages to kafka in
        message-by-message (send and flush for each message) mode.
        """
        return self.sends(topic, records=[record], binary=binary)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kafka_bridge_client']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'mypy-extensions>=0.4.3,<0.5.0',
 'requests>=2.27.1,<3.0.0']

setup_kwargs = {
    'name': 'kafka-bridge-client',
    'version': '0.7.0',
    'description': 'Python client for Strimzi Kafka Bridge',
    'long_description': "# kafka-bridge-client\nPython async client for [Strimzi Kafka Bridge](https://github.com/strimzi/strimzi-kafka-bridge) and [Confluent REST Proxy](https://docs.confluent.io/platform/current/kafka-rest/index.html) Package include consumer only.\n\n[![License: MIT](https://img.shields.io/badge/License-MIT-green.svg)](https://opensource.org/licenses/MIT)\n[![PRs Welcome](https://img.shields.io/badge/PRs-welcome-green.svg)](https://github.com/shafa-dev/kafka-bridge-client/issues)\n[![PyPI version](https://badge.fury.io/py/kafka-bridge-client.svg)](https://badge.fury.io/py/kafka-bridge-client)\n\n## Install\n```\npip install kafka-bridge-client\n```\n\n## Usage\nBy default client use [Strimzi Kafka Bridge](https://github.com/strimzi/strimzi-kafka-bridge) API\n\n`Consumer (async)`\n\n```python\nfrom kafka_bridge_client import KafkaBridgeConsumer\n\n# Strimzi Kafka Bridge\n\nconsumer1 = KafkaBridgeConsumer(\n    'topic1',\n    'topic2',\n    group_id='my-group,\n    auto_offset_reset='earliest',\n    enable_auto_commit=False,\n    bootstrap_server='your-kafka-bridge-url',\n    consumer_name='consumer-name',\n)\n\n# Confluent REST Proxy\nconsumer2 = KafkaBridgeConsumer(\n    'topic1',\n    'topic2',\n    group_id='my-group,\n    auto_offset_reset='earliest',\n    enable_auto_commit=False,\n    bootstrap_server='your-kafka-bridge-url',\n    consumer_name='consumer-name',\n    proxy='confluent'\n)\n\nasync for rec in consumer1.get_records():\n    print(rec['value'])\n    await consumer.commit()\n\n# or\n\nrecords = await consumer1.poll_records()\nprint(records)\nawait consumer.commit()\n```\n\n\n`Producer (sync)`\n\n```python\nfrom kafka_bridge_client import KafkaBridgeProducer\n\nproducer = KafkaBridgeProducer('http://bridge.url' timeout=5)\nproducer.send(Message(key='1', value='value'))\n```\n\n\n## Deploy\n\nYou need to change version in `pyproject.toml` and run it\n\n```\npoetry publish --build\n```",
    'author': 'Bogdan Zaseka',
    'author_email': 'zaseka.bogdan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/shafa-dev/kafka-bridge-client',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
